#property version   "1.0"
#property strict
#property copyright "(c)2021 Edgar Akhmadeev"
#property link      "https://www.mql5.com/en/users/dali"
// 2021.03.23



#include <WinAPI\sysinfoapi.mqh>



struct UUID {
	ulong	l;
	ulong	h;
};

union UUID16 {
	uchar	b[16];
	UUID	d;
};



#define SMBIOS_count		19
#define SMBIOS_fields	30



#define SIG(a, b, c, d)	((a << 24) + (b << 16) + (c << 8) + d)
//const uint	SMBIOS_signature = 'RSMB';
//const uint 	SMBIOS_signature = ('R' << 24) + ('S' << 16) + ('M' << 8) + 'B';
const uint 		SMBIOS_signature = SIG('R','S','M','B');

#ifdef SMBIOS_DEBUG
const string	SMBIOS_labels[SMBIOS_count] = {
	"System Management BIOS Version",
	
	"System BIOS Version",
	"Embedded Controller Firmware Version",
	
	"BIOS Vendor",
	"BIOS Version",
	"BIOS Release Date",
	
	"UUID",
	
	"System Manufacturer",
	"Product Name",
	"Product Version",
	"Product Serial Number",
	"SKU Number",
	"Family",
	
	"Mainboard Manufacturer",
	"Mainboard Name",
	"Mainboard Version",
	"Mainboard Serial Number",
	"Asset Tag",
	"Location in chassis"
};
#endif 



class SMBIOS {
private:
	static UUID16	uuid;
	
	#ifdef SMBIOS_DEBUG
	static string	info[SMBIOS_fields];
	#endif 
	
public:
	static string	Read();

	#ifdef SMBIOS_DEBUG
	static void		Print();
	
	static string	GetProductSN()	{ return info[10]; }
	static string	GetBoardSN()	{ return info[16]; }
	#endif 
};



static UUID16	SMBIOS::uuid;

#ifdef SMBIOS_DEBUG
static string	SMBIOS::info[SMBIOS_fields];
#endif 



static string
SMBIOS::Read() {
	// Query size of SMBIOS info.
	uint size = GetSystemFirmwareTable(SMBIOS_signature, 0, NULL, 0);
	
	// Allocate memory for SMBIOS info
	uchar data[];
	if (ArrayResize(data, size) <= 0)
		return NULL;
	
	// Retrieve the SMBIOS table
	uint written = GetSystemFirmwareTable(SMBIOS_signature, 0, data, size);
	if (written != size)
		return NULL;
	
	uint idx = 0;
	
	#ifdef SMBIOS_DEBUG
	int cnt = 0;
	info[cnt++] = StringFormat("%d.%d", data[idx + 1], data[idx + 2]);
	#endif 
	
	idx += 8;
	
	// Process the SMBIOS info
	
	string sUUID;
	
	while (idx < size) {
		uchar type = data[idx];
		uchar len = data[idx + 1];
		if (len < 4)
			break;
		
		if (type == 1 && len >= 25) {
			idx += 8;	// offset to UUID
			
			// check if there is a valid UUID (not all 0x00 or all 0xff)
			bool all_zero = true, all_one = true;
			for (int i = 0; i < 16 && (all_zero || all_one); i++) {
				if (data[idx + i] != 0x00)
					all_zero = false;
				if (data[idx + i] != 0xFF)
					all_one = false;
			}
		
			if (!all_zero && !all_one) {
				uint idx2 = 0;
				sUUID = "{";
				// As off version 2.6 of the SMBIOS specification, the first 3 fields of the UUID are supposed to be encoded on little-endian. (para 7.2.1)
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 3]);
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 2]);
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 1]);
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 0]);
				sUUID += "-";
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 5]);
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 4]);
				sUUID += "-";
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 7]);
				sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + 6]);
				sUUID += "-";
				for (int i = 8; i < 10; i++)
					sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + i]);
				sUUID += "-";
				for (int i = 10; i < 16; i++)
					sUUID += StringFormat("%02x", uuid.b[idx2++] = data[idx + i]);
				sUUID += "}";
				
				#ifdef SMBIOS_DEBUG
				info[cnt] = sUUID;
				#endif 
			}
			
			#ifdef SMBIOS_DEBUG
			++cnt;
			idx += 16;		// Skip uuid bytes
			idx += 3;		// Skip unknown info
			
			#else
			return sUUID;
			#endif 
		}
		
		#ifdef SMBIOS_DEBUG
		else if (type == 0) {
			if ((data[idx + 20] != 0 || data[idx + 21] != 0) && (data[idx + 20] != 255 || data[idx + 21] != 255))
				info[cnt] = StringFormat("%d.%d", data[idx + 20], data[idx + 21]);
			++cnt;
			if ((data[idx + 22] != 0 || data[idx + 23] != 0) && (data[idx + 22] != 255 || data[idx + 23] != 255))
				info[cnt] = StringFormat("%d.%d", data[idx + 22], data[idx + 23]);
			++cnt;
			idx += len;
		}
		#endif 
		
		else
			idx += len;		// Skip formatted area
		
		#ifdef SMBIOS_DEBUG
		if (cnt >= SMBIOS_count)
			break;			// Finish after main info
		#endif 
		
		do {
			int i = 0;
			while (idx + i < size && data[idx + i] != 0)
				++i;
			
			#ifdef SMBIOS_DEBUG
			info[cnt] = CharArrayToString(data, idx, i);
			if (info[cnt] == "" || info[cnt] == "To be filled by O.E.M.")
				info[cnt] = NULL;
			++cnt;
			#endif 
			
			idx += i + 1;
		} while (data[idx] != 0);
		++idx;
	}
	
	return sUUID;
}



#ifdef SMBIOS_DEBUG
static void
SMBIOS::Print() {
	int size = ArraySize(SMBIOS_labels);
	
	//for (int i = 0; i < size; ++i)
	//	::Print(SMBIOS_labels[i], ": ", info[i]);
	
	for (int i = 0; i < SMBIOS_fields; ++i) {
		if (info[i] != NULL) {
			if (i < size)
				::Print(SMBIOS_labels[i], ": ", info[i]);
			else
				::Print(info[i]);
		}
	}
}
#endif 
